import socket, select, regex, os, string, regsub, copy
import mud, app, player


# Close all open connections
def ClearAll():
    for ses in sessionlist:
	ses.Close()

# retrieve session by name
# return None if unfound
def GetSes(str):
    if str == 'all':
	return 'all'
    for ses in sessionlist:
	if ses.name == str:
	    return ses
    return None

# the default data hook: see class
# databuffer
def default_data_hook(dbuffer):
    pass


##################################################################
# session class
# specific connection to a mud
##################################################################

class session:
    def __init__(self, name, domain, port):
	self.aliases = {}
	self.actions = {}
	self.subs = {}
	self.vars = {}
	self.gags = []
	# we save a certain amount of previous mud output in the 'databuffer'
	# for later inspection by python functions
	self.databuf = databuffer()
	self.connected = 0
	self.name = ''
	self.logfile = None
	self.logging = 0
	self.sorck = None
	self.speedwalk = 1
	if name:
	    self.name = name
	if domain and port:
	    self.sorck= socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	    self.sorck.connect(domain, port)
	    self.name = name
	    self.connected = 1

    # write anything from our connection to stdout
    def ReadMud(self):
	try:
	    # check if there's anything to read
	    if not self.connected: 
		return ''
	    readers,w,e = select.select([self.sorck], [], [], timeout)
	    if not readers: # timer expired
		return '' 
	    for R in readers:
		data = R.recv(BUFSIZE)
		if not data: 
		    return ''
		return data 
	except socket.error, x:
	    self.Die(x)
    
    # initialize a new local session.
    # inherits all its aliases etc from the common session
    def InitLocalSession(self):
	self.aliases = copy.copy(common.aliases)
	self.actions = copy.copy(common.actions)
	self.gags = copy.copy(common.gags)
	self.subs = copy.copy(common.subs)	

    # die
    def Die(self, exc):
	global common, numsessions, sessionlist, currsession
	import types
	if type(exc) == types.StringType:
	    player.Putline(exc)
	else:
	    player.Putline(exc[1])
	self.Close()
	player.Putline('session "' + self.name + '" died.')
	numsessions = numsessions - 1
	sessionlist.remove(self)
	# try to switch to another live session
	if len(sessionlist) > 1:
	    currsession = sessionlist[1]
	    player.Putline('session "' + currsession.name + '" activated.')
	else:
	    # no more live sessions
	    currsession = common
	    player.Putline('no more active sessions')
	    player.Prompt()
	    
    # write text to session's log file, but only if logging is turned on
    def log(self, text):
	if self.logging:
	    self.logfile.write(text)

    # see if text contains anything in our gag list
    def CheckForGaggedText(self, text):
	if text:
	    for gt in self.gags:
		if regex.search(gt, text) != -1:
		    return 1
	    return 0

    # connect to a mud
    def Connect(domain, port):
	try:
	    if not self.domain or not self.port:
		return
	    self.sorck = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	    self.sorck.connect(domain, port)
	    self.connected = 1
	except:
	    player.Putline('unable to connect')

    # write anything from stdin to our connection
    def WriteTo(self, data):
	mud.log(data)
	self.log(data)
	if not self.connected:
	    print "data.WriteTo: ** Internal Error: not connected  **"
	    return
	try:
	    self.sorck.send(data)
	except socket.error, x:
	    self.Die(x)

    # close the instance's socket
    def Close(self):
	if self.connected:
	    try:
		self.sorck.close()
	    except socket.error:
		pass
    
    # input a string, return a list of all the aliases that match it
    def ExpandAlias(self, str):
	return self.Expand(str, self.aliases.keys())

    # input a string, return a list of all the action triggers that match it
    def ExpandAction(self, str):
	return self.Expand(str, self.actions.keys())

    # input a string and a dict, return a list of all the dict's keys that
    # match the string
    def Expand(self, s, list):
	str = s[:]
	ret = []
	wildcard = string.count(str, '*')
	# convert '*' to '.*'
	str = regsub.gsub('\*', '.*', str)
	for s in list:
	    if wildcard:
		if regex.match(str, s) != -1:
		    ret = ret + [s]
	    else:
		if str == s:
		    ret = ret + [s]
	return ret

    # wipe session clean of actions/aliases/subs/gags
    def Clear(self):
	self.aliases = {}
	self.actions = {}
	self.subs = {}
	self.gags = []
	self.vars = {}

    # see if text is either a variable or the beginning of a variable
    def Isvar(self, text):
	for v in self.vars.keys():
	    if string.find(v, text) == 0:
		return 1
	return 0
	
    # see if text is a variable
    def IsRealvar(self, text):
	for v in self.vars.keys():
	    if v == text:
		return 1
	return 0
	
    # see if any mud output triggers an action.
    def CheckActions(self, output):
	for ses in [self]:
	    for ac in ses.actions.keys():
		# convert lyntin vars into regices
		# if speed becomes a problem, we could get around this...		
		flac = regsub.gsub('%[0-9]+', '\(.*\)', ac)
		# code to support triggers anchored to beginning of line
		anchored = 0
		if flac[0] == '^':
		    anchored = 1
		    flac = flac[1:]
		matched = []
		ind = regex.search(flac, output) 
#		print 'searching for', flac
		if ind != -1:
#		    print 'found a match'
		    # split input into lines
		    lines = string.splitfields(output, '\r')
		    if anchored:
			for line in lines:
			    # strip garbage from the beginning (wtf is this?)
			    line = regsub.gsub('^%s[[0-9]+m'%chr(27), '', line)
			    # match must be at the beginning
			    if regex.match(flac, line) != -1:
				line = regsub.sub('^[[0-9][0-9]', '', line)
				matched.append(line)
		    else:
			for line in lines:
			    # match can be anywhere
			    if regex.search(flac, line) != -1:
				line = regsub.sub('^[[0-9][0-9]', '', line)
				matched.append(line)
		    flac_regex = regex.compile(flac)
		    for match in matched:
			# register the backreferences ('group' method on regex objects)
			flac_regex.search(match)
			schwag = ses.actions[ac][:]
			count = string.count(flac, '\(.*\)')
			# build _ordered_ list of variables to fill
			order = []
			ret = schwag[:]
			while var_regex.search(ret) != -1:
			    num = string.atoi(var_regex.group(1))
			    order = order + [num]
			    ret = regsub.sub('%\([0-9]+\)', '', ret)
#			    print 'ret becomes', ret
			# fill in response variables from those that matched on the trigger
			i = 0
			while i < count:
			    bref = flac_regex.group(order[i])
			    bref = regsub.gsub(';', '', bref)
#			    print 'i is %s, bref is %s'%(i, bref)
			    if bref[-1] == '\n' or bref[-1] == '\r':
				# this avoids a nasty bug, and I DOUBT anyone will
				# need newlines in their variables
				bref = bref[:-1]
			    schwag = regsub.sub('%' + str(order[i]), bref, schwag)
			    i = i + 1
			# make my day
			theapp.HandleUserInput(schwag)

##################################################################
# databuffer class
# class for storing previous output from the mud.
# mainly for inspection from python functions, but 
# grepping methods are also provided.
# anytime text is added to the databuffer, the member 'data_hook',
# which should be a callable object of one argument of type databuffer,
# is called.
##################################################################

class databuffer:
    def __init__(self, size=20):
	self.size = size
	self.list = []
	self.data_hook = default_data_hook

    # add a chunk of text
    def add(self, text):
	self.list = [text] + self.list
	# chop the storage list if it's too big
	if len(self.list) > self.size:
	    while len(self.list) > self.size:
		del self.list[-1]
	# run the hook, which does nothing by default
	self.run_hook()
    
    # resize the buffer
    def resize(self, s):
	if s < 1:
	    raise 'LTDatabufferError', 'can\'t set buffer size to less than 1'
	self.size = s

    # search through the databuffer for a regular expression.
    # return a list of all the entries that matched it.
    def grep(self, pat):
	ret = []
	for l in self.list:
	    if regex.search(pat, l) != -1:
		ret.append(l)
	return ret

    # search through the databuffer for a regular expression.
    # return a list of all the _lines_ that matched it.
    def greplines(self, pat):
	build = []
	for g in self.list:
	    lines = string.splitfields(g, '\r')
	    for line in lines:
		if regex.search(pat, line) != -1:
		    build.append(line)
	return build

    # run the hook. this is for customization and agent creation
    def run_hook(self):
	return self.data_hook(self)

##################################################################
# Global Variables 
##################################################################

# The application
theapp = None

# base directory to look for lyntin files in
if os.environ.has_key('LYNTINDIR'):
    initdir = os.environ['LYNTINDIR']
else:
    initdir = os.getcwd()
if initdir[-1] != '/':
    initdir = initdir + '/'

# current number of nonglobal sessions
numsessions = 0

# log file
logfile = open(initdir + 'mudlog', 'w')

# how long we'll wait on a socket for new data
timeout = .01

# common aliases/actions and such
common = session('common', None, None)

# the active session
currsession = common

# global session list
sessionlist = [common]

# how much data we'll read at once
BUFSIZE = 4096

# what a variable looks like
var_regex = regex.compile('%\([0-9]+\)') 

# latest commands mudder typed
history = []

# size of history
histsize = 30
